"""
MIT License

Copyright (c) 2019 NinjaSnail1080
Copyright (c) 2022 avizum

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import json
import re
import time
from typing import Any

import aiohttp

from .exceptions import CanNotGoBack, NotStarted
from .utils import MISSING, raise_connection_error
from .models import Answer, Language, Theme, Guess


NEW_SESSION_URL = "https://{}/new_session?callback=jQuery331023608747682107778_{}&urlApiWs={}&partner=1&childMod={}&player=website-desktop&uid_ext_session={}&frontaddr={}&constraint=ETAT<>'AV'&soft_constraint={}&question_filter={}"
ANSWER_URL = "https://{}/answer_api?callback=jQuery331023608747682107778_{}&urlApiWs={}&childMod={}&session={}&signature={}&step={}&answer={}&frontaddr={}&question_filter={}"
BACK_URL = "{}/cancel_answer?callback=jQuery331023608747682107778_{}&childMod={}&session={}&signature={}&step={}&answer=-1&question_filter={}"
WIN_URL = "{}/list?callback=jQuery331023608747682107778_{}&childMod={}&session={}&signature={}&step={}"


HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "en-US,en;q=0.9",
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) snap Chromium/81.0.4044.92 Chrome/81.0.4044.92 Safari/537.36",
    "x-requested-with": "XMLHttpRequest",
}


class Akinator:
    """
    A class that represents an async Akinator game.

    .. note::

        Some attributes will be missing before a game has started.
        Use :func:`Akinator.start` a game before accessing these attributes.
        Some attributes will be missing until a game has ended.
        Use :func:`Akinator.win` to end a game before accessing these attributes.

    Parameters
    ----------
    language: :class:`Language`
        The language to use when starting the game. If left blank, defaults to :attr:`Language.ENGLISH`.
    theme: :class:`Theme`
        The theme to use when starting the game. If left blank, defaults to :attr:`Theme.CHARACTERS`.
    child_mode: :class:`bool`
        Whether to use child mode or not. Defaults to False.

    Attributes
    ----------
    question: :class:`str`
        The question that Akinator is asking.
    progression: :class:`float`
        How far in the game you are.
    step: :class:`int`
        The question you are on, starting from 0.
    first_guess: :class:`Guess`
        A dictionary containing the first guess information.
    guesses: list[:class:`Guess`]
        A list of :class:`Guess` dictionary of guesses from greatest to least probability.

    uri: :class:`str`
        The uri that is being used.
    server: :class:`str`
        The server that is being used.
    signature: :class:`int`
        An :class:`int` that represents a game's signature.
    uid: :class:`str`
        Represents a games Unique ID, used for authentication purposes.
    frontaddr: :class:`str`
        An IP address in Base64, used for authentication purposes.
    timestamp: :class:`int`
        A POSIX timestamp that is set when :meth:`Akinator.start` is called
    session: :class:`int`
        Represents a game's session.

    Raises
    ------
    :exc:`TypeError`
        The session is not an :class:`aiohttp.ClientSession`.
    """

    def __init__(
        self,
        language: Language = Language.ENGLISH,
        theme: Theme = Theme.CHARACTERS,
        child_mode: bool = False,
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        if session is not None and not isinstance(session, aiohttp.ClientSession):
            raise TypeError("session must be a aiohttp.ClientSession")

        self.language: Language = language
        self.theme: Theme = theme
        self.child_mode: bool = child_mode
        self._session: aiohttp.ClientSession = MISSING
        self._started: bool = False

        self.question: str = MISSING
        self.progression: float = 0.0
        self.step: int = 0

        self.first_guess: Guess = MISSING
        self.guesses: list[Guess] = MISSING

        self.uri: str = MISSING
        self.server: str = MISSING
        self.signature: int = MISSING
        self.uid: str = MISSING
        self.frontaddr: str = MISSING
        self.question_filter: str = MISSING
        self.timestamp: float = 0.0
        self.session: int = 0


    async def _create_session(self) -> None:
        self._session = aiohttp.ClientSession()
        self._session.headers.update(HEADERS)

    def _update(self, resp: Any) -> None:
        """Update class variables"""

        if not self._started:
            self.session = int(resp["parameters"]["identification"]["session"])
            self.signature = int(resp["parameters"]["identification"]["signature"])
            self.question = str(resp["parameters"]["step_information"]["question"])
            self.progression = float(resp["parameters"]["step_information"]["progression"])
            self.step = int(resp["parameters"]["step_information"]["step"])
        else:
            self.question = str(resp["parameters"]["question"])
            self.progression = float(resp["parameters"]["progression"])
            self.step = int(resp["parameters"]["step"])

    def _parse_response(self, response: Any) -> Any:
        """Parse the JSON response and turn it into a Python object"""

        return json.loads(",".join(response.split("(")[1::])[:-1])

    async def _get_session_info(self) -> None:
        """Get uid and frontaddr from akinator.com/game"""

        info_regex = re.compile("var uid_ext_session = '(.*)'\\;\\n.*var frontaddr = '(.*)'\\;")

        async with self._session.get("https://en.akinator.com/game") as w:
            match = info_regex.search(await w.text())
            if not match:
                return

        self.uid, self.frontaddr = match.groups()[0], match.groups()[1]

    async def _auto_get_region(self, language: Language, theme: Theme) -> dict[str, str]:
        """Automatically get the uri and server from akinator.com for the specified language and theme"""

        server_regex = re.compile(
            '[{"translated_theme_name":"[\s\S]*","urlWs":"https:\\\/\\\/srv[0-9]+\.akinator\.com:[0-9]+\\\/ws","subject_id":"[0-9]+"}]'
        )

        uri = f"{language}.akinator.com"
        default_return = {"uri": uri, "server": MISSING}
        async with self._session.get(f"https://{uri}") as w:
            match = server_regex.search(await w.text())
            if match is None:
                return default_return
        parsed = json.loads(match.group().split("'arrUrlThemesToPlay', ")[-1])
        server = MISSING
        server = next((i for i in parsed if i["subject_id"] == str(theme.value)), MISSING)["urlWs"]

        if server not in ["https://srv12.akinator.com:9398/ws"]:
            return {"uri": uri, "server": server}
        return default_return

    async def start(
        self,
        language: Language | None = None,
        theme: Theme | None = None,
        child_mode: bool | None = None,
    ) -> str:
        """
        Starts a new game. This should be called before any other method.


        Parameters
        ----------
        language: :class:`Language` | :class:`None`
            The language to use when starting the game. If :class:`None`, the language specified in the constructor will be used.
        theme: :class:`Theme` | :class:`None`
            The theme to use when starting the game. If :class:`None`, the theme specified in the constructor will be used.
        child_mode: :class:`bool`
            Whether or not to use child mode. If True, the game will be more "child-friendly". If :class:`None`, the child mode specified in the constructor will be used.

        Returns
        -------
        :class:`str`
            The first question that Akinator is asking.
        """
        if self._session is MISSING:
            await self._create_session()

        if language is not None:
            self.language = language
        if theme is not None:
            self.theme = theme
        if child_mode is not None:
            self.child_mode = child_mode

        self.timestamp = time.time()

        region_info = await self._auto_get_region(self.language, self.theme)
        self.uri, self.server = region_info["uri"], region_info["server"]

        soft_constraint = "ETAT%3D%27EN%27" if self.child_mode else ""
        self.question_filter = "cat%3D1" if self.child_mode else ""
        await self._get_session_info()

        async with self._session.get(
            NEW_SESSION_URL.format(
                self.uri,
                self.timestamp,
                self.server,
                str(self.child_mode).lower(),
                self.uid,
                self.frontaddr,
                soft_constraint,
                self.question_filter,
            ),
            headers=HEADERS,
        ) as w:
            resp = self._parse_response(await w.text())

        if resp["completion"] == "OK":
            self._started = True
            self._update(resp)
            return self.question
        else:
            return raise_connection_error(resp["completion"])

    async def answer(self, answer: Answer) -> str:
        """
        Answers the current question accessed with :attr:`Akinator.question`, and returns the next question.

        Parameter
        ---------
        answer: :class:`Answer`
            The answer to the current question.

            .. note::

                Calling this method with :attr:`Answer.BACK` will call :meth:`Akinator.back` for you.

        Raises
        ------
        :exc:`NoMoreQuestions`
            There are no more questions to be asked. Call :meth:`Akinator.win` to get the results.
        :exc:`CanNotGoBack`
            The Akinator game is on the first question, so it can't go back anymore.
        :exc:`NotStarted`
            Occurs when you try to answer a question before starting the game.

        Returns
        -------
        :class:`str`
            The next question that Akinator asks.
        """

        if self._started is False:
            raise NotStarted("Game has not started!")

        if answer == Answer.BACK:
            return await self.back()

        async with self._session.get(
            ANSWER_URL.format(
                self.uri,
                self.timestamp,
                self.server,
                str(self.child_mode).lower(),
                self.session,
                self.signature,
                self.step,
                str(answer.value),
                self.frontaddr,
                self.question_filter,
            ),
            headers=HEADERS,
        ) as w:
            resp = self._parse_response(await w.text())

        if resp["completion"] == "OK":
            self._update(resp)
            return self.question
        else:
            return raise_connection_error(resp["completion"])

    async def back(self) -> str:
        """
        Go back to the previous question.

        .. note::

            :meth:`Akinator.answer` will call this method for you if you pass in :attr:`Answer.BACK`.
            It is recommended that you use that for better experience.

        Raises
        ------
        :exc:`CanNotGoBack`
            The Akinator game is on the first question, so it can't go back anymore.
        :exc:`NotStarted`
            Occurs when you try to answer a question before starting the game.

        Returns
        -------
        :class:`str`
            The previous question that Akinator asked.
        """

        if self._started is False:
            raise NotStarted("Game has not started!")

        if self.step == 0:
            raise CanNotGoBack("You were on the first question and couldn't go back any further")

        async with self._session.get(
            BACK_URL.format(
                self.server,
                self.timestamp,
                str(self.child_mode).lower(),
                self.session,
                self.signature,
                self.step,
                self.question_filter,
            ),
            headers=HEADERS,
        ) as w:
            resp = self._parse_response(await w.text())

        if resp["completion"] == "OK":
            self._update(resp)
            return self.question
        else:
            return raise_connection_error(resp["completion"])

    async def win(self) -> Guess:
        """
        Get Akinator's current guesses based on the responses to the questions thus far.

        This function will set:

        :attr:`Akinator.first_guess`
            The first guess that is returned
        :attr:`Akinator.guesses`
            A list of guesses from greatest to lowest probablity

        .. note::

            It is recommended that you call this function when `Akinator.progression` is above 85.0,
            because by then, Akinator will most likely narrowed the guesses down to one.

        Raises
        ------
        :exc:`NotStarted`
            Occurs when you try to answer a question before starting the game.

        Returns
        -------
        :class:`Guess`
            The first guess that Akinator has. Also set in :attr:`Akinator.first_guess`.
        """

        if self._started is False:
            raise NotStarted("Game has not started!")

        async with self._session.get(
            WIN_URL.format(
                self.server,
                self.timestamp,
                str(self.child_mode).lower(),
                self.session,
                self.signature,
                self.step,
            ),
            headers=HEADERS,
        ) as w:
            resp = self._parse_response(await w.text())

        if resp["completion"] == "OK":
            self.first_guess = Guess._from_dict(resp["parameters"]["elements"][0]["element"])
            self.guesses = [Guess._from_dict(x["element"]) for x in resp["parameters"]["elements"]]
            return self.first_guess
        else:
            return raise_connection_error(resp["completion"])

    async def end(self) -> Guess:
        """
        Alias to :meth:`Akinator.win`.
        """
        return await self.win()

    async def close(self) -> None:
        """
        Closes the aiohttp ClientSession.

        .. caution::

            If you specified your own ClientSession, this may interrupt what you are doing with the session.
        """
        if self._session is not MISSING and self._session.closed is False:
            await self._session.close()

        self._session = MISSING
        self._started = False



"""
MIT License

Copyright (c) 2019 NinjaSnail1080
Copyright (c) 2022 avizum

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class InputError(ValueError):
    """Raised when the user inputs an invalid answer"""

    pass


class InvalidAnswer(InputError):
    """Raised when the user inputs an invalid answer"""

    pass


class InvalidLanguage(InputError):
    """Raised when the user inputs an invalid language"""

    pass


class InvalidTheme(InputError):
    """Raised when the user inputs an invalid theme"""

    pass


class ConnectionFailure(Exception):
    """Raised if the Akinator API fails to connect for some reason."""

    pass


class TimedOut(ConnectionFailure):
    """Raised if the Akinator session times out"""

    pass


class NoMoreQuestions(ConnectionFailure):
    """Raised if the Akinator API runs out of questions to ask. This will happen if "Akinator.step" is at 79"""

    pass


class ServerDown(ConnectionFailure):
    """Raised if Akinator's servers are down for the region you're running on."""

    pass


class TechnicalServerError(ConnectionFailure):
    """Raised if Akinator's servers had a technical error."""

    pass


class NotStarted(Exception):
    """Raised when the user tries to do something before starting the game."""

    pass


class CanNotGoBack(Exception):
    """Raised when the user is on the first question and tries to go back further."""

    pass



"""
MIT License

Copyright (c) 2019 NinjaSnail1080

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import Any, NoReturn


from .exceptions import (
    ConnectionFailure,
    TimedOut,
    NoMoreQuestions,
    ServerDown,
    TechnicalServerError,
)


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self):
        return "..."


MISSING: Any = _MissingSentinel()


def raise_connection_error(response: str) -> NoReturn:
    """Raise the proper error if the API failed to connect"""
    if response == "KO - SERVER DOWN":
        raise ServerDown("Akinator's servers are down in this region. Try again later or use a different language")

    elif response == "KO - TECHNICAL ERROR":
        raise TechnicalServerError(
            "Akinator's servers have had a technical error. Try again later or use a different language"
        )

    elif response == "KO - TIMEOUT":
        raise TimedOut("Your Akinator session has timed out")
    elif response in {"KO - ELEM LIST IS EMPTY", "WARN - NO QUESTION"}:
        raise NoMoreQuestions('"Akinator.step" reached 79. No more questions.')
    else:
        raise ConnectionFailure(f"An unknown error has occured. Server response: {response}")



"""
MIT License

Copyright (c) 2022 avizum

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import enum
import dataclasses
from .exceptions import InvalidAnswer, InvalidLanguage, InvalidTheme
from typing import TypedDict


__all__ = (
    "Answer",
    "Language",
    "Theme",
    "Guess",
)


class Answer(enum.Enum):
    """
    Represents the answer to a question.
    """

    YES = 0
    NO = 1
    I_DONT_KNOW = 2
    PROBABLY = 3
    PROBABLY_NOT = 4
    BACK = 5

    def __str__(self) -> str:
        return self.name

    def __int__(self) -> int:
        return self.value

    def __repr__(self) -> str:
        return f"<Answer answer={self.name}>"

    @classmethod
    def from_str(cls, answer: str) -> Answer:
        """
        Allows for the creation of an Answer from a string.


        Parameters
        ----------
        answer: :class:`str`
            The answer to convert to :class:`Answer`.

            For :attr:`Answer.YES`, the following are valid:
                ``yes``, ``y``, and ``0``.
            For :attr:`Answer.NO`, the following are valid:
                ``no``, ``n``, and ``1``.
            For :attr:`Answer.I_DONT_KNOW`, the following are valid:
                ``i dont know``, ``i don't know``, ``idk``, and ``2``.
            For :attr:`Answer.PROBABLY`, the following are valid:
                ``probably``, ``p``, and ``3``.
            For :attr:`Answer.PROBABLY_NOT`, the following are valid:
                ``probably not``, ``pn``, and ``4``.
            For :attr:`Answer.BACK`, the following are valid:
                ``back``, ``b``, and ``5``.

        Raises
        ------
        :exc:`InvalidAnswer`
            The answer you provided is not valid.

        Returns
        -------
        :class:`Answer`
        """

        answer = answer.lower()
        if answer in {"yes", "y", "0"}:
            return cls.YES
        elif answer in {"no", "n", "1"}:
            return cls.NO
        elif answer in {"i", "idk", "i dont know", "i don't know", "2"}:
            return cls.I_DONT_KNOW
        elif answer in {"probably", "p", "3"}:
            return cls.PROBABLY
        elif answer in {"probably not", "pn", "4"}:
            return cls.PROBABLY_NOT
        elif answer in {"back", "b", "5"}:
            return cls.BACK
        else:
            raise InvalidAnswer(f"Invalid answer: {answer}")


class Language(enum.Enum):
    """
    Represents the language of the game.
    """

    ENGLISH = "en"
    ARABIC = "ar"
    CHINESE = "cn"
    GERMAN = "de"
    SPANISH = "es"
    FRENCH = "fr"
    HEBREW = "il"
    ITALIAN = "it"
    JAPANESE = "jp"
    KOREAN = "kr"
    DUTCH = "nl"
    POLISH = "pl"
    PORTUGUESE = "pt"
    RUSSIAN = "ru"
    TURKISH = "tr"
    INDONESIAN = "id"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"<Language language={self.value}>"

    @classmethod
    def from_str(cls, language: str) -> Language:
        """
        Create a :class:`Language` from a string.

        The short form of a language (``en`` for English, ``fr`` for French, etc.) is also accepted.

        Parameters
        ----------
        language: :class:`str`
            The language to convert to :class:`Language`.

        Raises
        ------
        :exc:`InvalidLanguage`
            The language you provided is not valid.

        Returns
        -------
        :class:`Language`
        """
        language = language.lower()
        if language in {"english", "en"}:
            return cls.ENGLISH
        elif language in {"arabic", "ar"}:
            return cls.ARABIC
        elif language in {"chinese", "cn"}:
            return cls.CHINESE
        elif language in {"german", "de"}:
            return cls.GERMAN
        elif language in {"spanish", "es"}:
            return cls.SPANISH
        elif language in {"french", "fr"}:
            return cls.FRENCH
        elif language in {"hebrew", "il"}:
            return cls.HEBREW
        elif language in {"italian", "it"}:
            return cls.ITALIAN
        elif language in {"japanese", "jp"}:
            return cls.JAPANESE
        elif language in {"korean", "kr"}:
            return cls.KOREAN
        elif language in {"dutch", "nl"}:
            return cls.DUTCH
        elif language in {"polish", "pl"}:
            return cls.POLISH
        elif language in {"portuguese", "pt"}:
            return cls.PORTUGUESE
        elif language in {"russian", "ru"}:
            return cls.RUSSIAN
        elif language in {"turkish", "tr"}:
            return cls.TURKISH
        elif language in {"indonesian", "id"}:
            return cls.INDONESIAN
        else:
            raise InvalidLanguage(f"Invalid language: {language}")


class Theme(enum.Enum):
    """
    Determines what server to use when starting a game.
    """

    CHARACTERS = 1
    OBJECTS = 2
    ANIMALS = 14

    @classmethod
    def from_str(cls, theme: str) -> Theme:
        """
        Create a :class:`Theme` from a string.

        Parameters
        ----------
        theme: :class:`str`
            The theme to convert to :class:`Theme`.

            For :attr:`Theme.CHARACTERS`, the following are valid:
                ``characters``, ``ch``, and ``c``.
            For :attr:`Theme.OBJECTS`, the following are valid:
                ``objects``, ``obj``, and ``o``.
            For :attr:`Theme.ANIMALS`, the following are valid:
                ``animals``, and ``a``.

        Raises
        ------
        :exc:`InvalidTheme`
            The theme you provided is not valid.

        Returns
        -------
        :class:`Theme`
        """
        theme = theme.lower()
        if theme in {"characters", "ch", "c"}:
            return cls.CHARACTERS
        elif theme in {"objects", "obj", "o"}:
            return cls.OBJECTS
        elif theme in {"animals", "a"}:
            return cls.ANIMALS
        else:
            raise InvalidTheme(f"Invalid theme: {theme}")


class _GuessDict(TypedDict):
    id: int
    name: str
    id_base: int
    proba: float
    description: str
    valide_contrainte: int
    ranking: int
    pseudo: str
    picture_path: str
    corrupt: int
    relative: int
    award_id: int
    flag_photo: int
    absolute_picture_path: str


@dataclasses.dataclass
class Guess:
    """
    Class representing a guess.

    Attributes
    ----------
    id: :class:`int`
        The ID of the guess.
    name: :class:`str`
        The name of the guess.
    probablility: :class:`float`
        The probability that the guess is correct.
    description: :class:`str`
        The description of the guess.
    ranking: :class:`int`
        The ranking of the guess.
    absolute_picture_path: :class:`str`
        A URL to the picture of the guess.
    """

    id: int
    name: str
    probablility: float
    description: str
    ranking: int
    absolute_picture_path: str
    _data: _GuessDict | None

    @classmethod
    def _from_dict(cls, data: _GuessDict) -> Guess:
        return cls(
            data["id"],
            data["name"],
            data["proba"],
            data["description"],
            data["ranking"],
            data["absolute_picture_path"],
            data,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asyncakinator']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0']

setup_kwargs = {
    'name': 'asyncakinator',
    'version': '1.5.2',
    'description': 'An async API wrapper for Akinator, written in Python.',
    'long_description': 'asyncakinator\n=============\n\n\n.. image:: https://discord.com/api/guilds/751490725555994716/embed.png\n   :target: https://discord.gg/muTVFgDvKf\n   :alt: Support Server Invite\n\nAn async API wrapper for the online game, Akinator, written in Python.\n\n`Akinator <https://en.akinator.com/>`_ is a web-based game which tries to determine what character you are thinking of by asking a series of questions.\n\n\nInstalling\n----------\n\nTo install, just run the following command:\n\n.. code-block:: shell\n\n    # MacOS/Linux\n    python3 -m pip install -U asyncakinator\n\n    # Windows\n    py -3 -m pip install -U asyncakinator\n\n\nRequirements\n~~~~~~~~~~~~\n- Python ≥3.9\n\n- ``aiohttp``\n\n\nDocumentation\n-------------\nDocumention can be found `here. <https://asyncakinator.readthedocs.io/en/latest/>`_\n\n\nQuick Examples\n--------------\n\nHere\'s a quick little example of the library being used to make a simple, text-based Akinator game:\n\n.. code-block:: python\n\n    import asyncio\n\n    from asyncakinator import (\n        Akinator,\n        Answer,\n        CanNotGoBack,\n        InvalidAnswer,\n        Language,\n        NoMoreQuestions,\n        Theme\n    )\n\n\n    game = Akinator(\n        language=Language.ENGLISH,\n        theme=Theme.ANIMALS,\n    )\n\n\n    async def main():\n        question = await game.start()\n\n        while game.progression <= 80:\n            print(question)\n            user_input = input("Answer:  ")\n            try:\n                answer = Answer.from_str(user_input)\n            except InvalidAnswer:\n                print("Invalid answer")\n                continue\n            try:\n                question = await game.answer(answer)\n            except CanNotGoBack:\n                print("This is the first question, you can\'t go back.")\n                continue\n            except NoMoreQuestions:\n                break\n\n        await game.win()\n\n        correct = input(\n            f"You are thinking of {game.first_guess.name} ({game.first_guess.description}). "\n            f"Am I correct?\\n{game.first_guess.absolute_picture_path}\\n---\\nAnswer:  "\n        )\n        if Answer.from_str(correct) == Answer.YES:\n            print("Nice.")\n        else:\n            print("Maybe next time.")\n        await game.close()\n\n\n    asyncio.run(main())',
    'author': 'avizum',
    'author_email': 'juliusrt@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/avizum/asyncakinator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9',
}


setup(**setup_kwargs)
